@@ -1,6 +1,6 @@
 
 from .base import Event, Handler
-from .device import DeviceEvent, DeviceUpdate
+from .device import DeviceEvent, DeviceUpdate, FullUpdate
 from .handlers import handle_event, register_handler
 from .handlers import init as init_handlers, cleanup as cleanup_handlers
 from .port import PortEvent, PortAdd, PortRemove, PortUpdate, ValueChange
