@@ -120,6 +120,11 @@ STANDARD_ATTRDEFS = {
         'optional': True,
         'modifiable': True
     },
+    'internal': {
+        'type': 'boolean',
+        'optional': True,
+        'modifiable': True
+    },
     'virtual': {
         'type': 'boolean',
         'optional': True
@@ -144,8 +149,9 @@ class PortReadError(PortError):
 
 
 class InvalidAttributeValue(PortError):
-    def __init__(self, attr: str) -> None:
+    def __init__(self, attr: str, details: Optional[GenericJSONDict] = None) -> None:
         self.attr: str = attr
+        self.details: Optional[GenericJSONDict] = details
 
         super().__init__(attr)
 
@@ -231,13 +237,15 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
     def initialize(self) -> None:
         pass
 
-    def _get_attrdefs(self) -> AttributeDefinitions:
+    async def get_attrdefs(self) -> AttributeDefinitions:
         if self._attrdefs_cache is None:
             self._attrdefs_cache = dict(self.STANDARD_ATTRDEFS, **self.ADDITIONAL_ATTRDEFS)
 
-        return self._attrdefs_cache
+            # Don't expose units for boolean ports
+            if await self.get_type() == TYPE_BOOLEAN:
+                self._attrdefs_cache.pop('unit')
 
-    ATTRDEFS = property(_get_attrdefs)
+        return self._attrdefs_cache
 
     def invalidate_attrdefs(self) -> None:
         self._attrs_cache = {}
@@ -246,22 +254,26 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
         self._non_modifiable_attrs = None
         self._schema = None
 
-    def get_non_modifiable_attrs(self) -> Set[str]:
+    async def get_non_modifiable_attrs(self) -> Set[str]:
+        attrdefs = await self.get_attrdefs()
+
         if self._non_modifiable_attrs is None:
-            self._non_modifiable_attrs = set(n for (n, v) in self.ATTRDEFS.items() if not v.get('modifiable'))
+            self._non_modifiable_attrs = set(n for (n, v) in attrdefs.items() if not v.get('modifiable'))
 
         return self._non_modifiable_attrs
 
-    def get_modifiable_attrs(self) -> Set[str]:
+    async def get_modifiable_attrs(self) -> Set[str]:
+        attrdefs = await self.get_attrdefs()
+
         if self._modifiable_attrs is None:
-            self._modifiable_attrs = set(n for (n, v) in self.ATTRDEFS.items() if v.get('modifiable'))
+            self._modifiable_attrs = set(n for (n, v) in attrdefs.items() if v.get('modifiable'))
 
         return self._modifiable_attrs
 
     async def get_attrs(self) -> Attributes:
         d = {}
 
-        for name in self.ATTRDEFS.keys():
+        for name in await self.get_attrdefs():
             v = await self.get_attr(name)
             if v is None:
                 continue
@@ -341,7 +353,7 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
 
         # Skip an IO loop iteration, allowing setting multiple attributes before triggering a port-update
         await asyncio.sleep(0)
-        self.trigger_update()
+        await self.trigger_update()
 
     async def handle_attr_change(self, name: str, value: Attribute) -> None:
         method_name = f'handle_{name}'
@@ -372,6 +384,9 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
     async def is_persisted(self) -> bool:
         return await self.get_attr('persisted')
 
+    async def is_internal(self) -> bool:
+        return await self.get_attr('internal')
+
     def is_enabled(self) -> bool:
         return self._enabled
 
@@ -383,6 +398,14 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
         self._enabled = True
         self.invalidate_attr('enabled')
 
+        # Reset port expression
+        if self._expression:
+            sexpression = str(self._expression)
+            self.debug('resetting expression "%s"', sexpression)
+            self._expression = core_expressions.parse(self.get_id(), sexpression)
+
+            main.force_eval_expressions(self)
+
         try:
             await self.handle_enable()
 
@@ -465,15 +488,15 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
             self.debug('checking for expression circular dependencies')
             await core_expressions.check_loops(self, expression)
 
-        except core_expressions.ExpressionError as e:
+        except core_expressions.ExpressionParseError as e:
             self.error('failed to set expression "%s": %s', sexpression, e)
 
-            raise InvalidAttributeValue('expression') from e
+            raise InvalidAttributeValue('expression', details=e.to_json()) from e
 
         self.debug('setting expression "%s"', expression)
         self._expression = expression
 
-        main.force_eval_expressions()
+        main.force_eval_expressions(self)
 
     async def attr_get_transform_read(self) -> str:
         if self._transform_read:
@@ -497,15 +520,18 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
                     continue
 
                 if dep != f'${self._id}':
-                    raise core_expressions.ExpressionError('Transform expression depends other ports')
+                    raise core_expressions.ExternalDependency(
+                        port_id=dep[1:],
+                        pos=stransform_read.index(dep)
+                    )
 
             self.debug('setting read transform "%s"', transform_read)
             self._transform_read = transform_read
 
-        except core_expressions.ExpressionError as e:
+        except core_expressions.ExpressionParseError as e:
             self.error('failed to set transform read expression "%s": %s', stransform_read, e)
 
-            raise InvalidAttributeValue('transform_read') from e
+            raise InvalidAttributeValue('transform_read', details=e.to_json()) from e
 
     async def attr_get_transform_write(self) -> Optional[str]:
         if not await self.is_writable():
@@ -532,15 +558,18 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
                     continue
 
                 if dep != f'${self._id}':
-                    raise core_expressions.ExpressionError('Transform expression depends other ports')
+                    raise core_expressions.ExternalDependency(
+                        port_id=dep[1:],
+                        pos=stransform_write.index(dep)
+                    )
 
             self.debug('setting write transform "%s"', transform_write)
             self._transform_write = transform_write
 
-        except core_expressions.ExpressionError as e:
+        except core_expressions.ExpressionParseError as e:
             self.error('failed to set transform write expression "%s": %s', stransform_write, e)
 
-            raise InvalidAttributeValue('transform_write') from e
+            raise InvalidAttributeValue('transform_write', details=e.to_json()) from e
 
     @abc.abstractmethod
     async def read_value(self) -> NullablePortValue:
@@ -596,12 +625,6 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
     def push_value(self, value: PortValue, reason: str) -> None:
         asyncio.create_task(self.write_transformed_value(value, reason))
 
-    def has_pending_write(self) -> bool:
-        if self._write_value_queue.qsize() > 0:
-            return True
-
-        return self._pending_write
-
     async def _write_value_queued(self, value: PortValue, reason: str) -> None:
         done = asyncio.get_running_loop().create_future()
 
@@ -610,6 +633,7 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
                 self._write_value_queue.put_nowait((value, reason, done))
 
             except asyncio.QueueFull:
+                self.warning('write queue full, dropping oldest value')
                 self._write_value_queue.get_nowait()
 
             else:
@@ -795,7 +819,7 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
             d['value'] = None
 
         # Attributes
-        for name in self.get_modifiable_attrs():
+        for name in await self.get_modifiable_attrs():
             v = await self.get_attr(name)
             if v is None:
                 continue
@@ -805,12 +829,6 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
         return d
 
     async def cleanup(self) -> None:
-        # Cancel sequence
-        if self._sequence:
-            self.debug('canceling current sequence')
-            await self._sequence.cancel()
-            self._sequence = None
-
         self._write_value_task.cancel()
         await self._write_value_task
 
@@ -830,19 +848,19 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
             self.debug('removing persisted data')
             persist.remove(self.PERSIST_COLLECTION, filt={'id': self._id})
 
-        self.trigger_remove()
+        await self.trigger_remove()
 
-    def trigger_add(self) -> None:
-        core_events.handle_event(core_events.PortAdd(self))
+    async def trigger_add(self) -> None:
+        await core_events.handle_event(core_events.PortAdd(self))
 
-    def trigger_remove(self) -> None:
-        core_events.handle_event(core_events.PortRemove(self))
+    async def trigger_remove(self) -> None:
+        await core_events.handle_event(core_events.PortRemove(self))
 
-    def trigger_update(self) -> None:
-        core_events.handle_event(core_events.PortUpdate(self))
+    async def trigger_update(self) -> None:
+        await core_events.handle_event(core_events.PortUpdate(self))
 
-    def trigger_value_change(self) -> None:
-        core_events.handle_event(core_events.ValueChange(self))
+    async def trigger_value_change(self) -> None:
+        await core_events.handle_event(core_events.ValueChange(self))
 
     async def get_schema(self) -> GenericJSONDict:
         if self._schema is None:
@@ -852,7 +870,8 @@ class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
                 'additionalProperties': False
             }
 
-            for name, attrdef in self.ATTRDEFS.items():
+            attrdefs = await self.get_attrdefs()
+            for name, attrdef in attrdefs.items():
                 if not attrdef.get('modifiable'):
                     continue
 
@@ -922,14 +941,16 @@ class Port(BasePort, metaclass=abc.ABCMeta):
 
         self._tag: str = ''
         self._persisted: bool = False
+        self._internal: bool = False
 
 
-async def load(port_settings: List[Dict[str, Any]], raise_on_error: bool = True) -> List[BasePort]:
+async def load(port_args: List[Dict[str, Any]], raise_on_error: bool = True) -> List[BasePort]:
     port_driver_classes = {}
     ports = []
 
-    for port_spec in port_settings:
-        driver = port_spec.pop('driver', None)
+    # Create ports
+    for ps in port_args:
+        driver = ps.pop('driver', None)
         if not driver:
             if raise_on_error:
                 raise PortLoadError('Missing port driver')
@@ -958,10 +979,7 @@ async def load(port_settings: List[Dict[str, Any]], raise_on_error: bool = True)
         port_class_desc = f'{port_class.__module__}.{port_class.__name__}'
 
         try:
-            value = port_spec.pop('value', None)  # Initial value
-            port = port_class(**port_spec)
-            if value is not None:
-                port.set_value(value)
+            port = port_class(**ps)
 
             _ports_by_id[port.get_id()] = port
             ports.append(port)
@@ -974,7 +992,13 @@ async def load(port_settings: List[Dict[str, Any]], raise_on_error: bool = True)
 
             logger.error('failed to initialize port from driver %s: %s', port_class_desc, e, exc_info=True)
 
-    for old_id, new_id in settings.port_mappings.items():
+    # Map IDs
+    for port in ports:
+        old_id = port.get_id()
+        new_id = settings.port_mappings.get(old_id)
+        if not new_id:
+            continue
+
         if new_id in _ports_by_id:
             if raise_on_error:
                 raise PortLoadError('Cannot map port {old_id} to {new_id}: new id already exists')
@@ -1002,6 +1026,7 @@ async def load(port_settings: List[Dict[str, Any]], raise_on_error: bool = True)
         _ports_by_id.pop(old_id)
         _ports_by_id[port.get_id()] = port
 
+    # Load created ports
     for port in ports:
         try:
             await port.load()
@@ -1012,14 +1037,14 @@ async def load(port_settings: List[Dict[str, Any]], raise_on_error: bool = True)
 
             port.error('failed to load: %s', port, e, exc_info=True)
 
-        port.trigger_add()
+        await port.trigger_add()
 
     return ports
 
 
 async def load_one(cls: Union[str, type], args: Dict[str, Any]) -> BasePort:
-    port_settings = [dict(driver=cls, **args)]
-    ports = await load(port_settings, raise_on_error=True)
+    port_args = [dict(driver=cls, **args)]
+    ports = await load(port_args, raise_on_error=True)
 
     return ports[0]
 
@@ -1033,9 +1058,15 @@ def all_ports() -> Iterable[BasePort]:
 
 
 async def cleanup() -> None:
-    for port in _ports_by_id.values():
+
+    async def cleanup_port(port: BasePort) -> None:
+        await port.disable()
         await port.cleanup()
 
+    tasks = [cleanup_port(port) for port in _ports_by_id.values()]
+    if tasks:
+        await asyncio.wait(tasks)
+
 
 def reset() -> None:
     logger.debug('clearing ports persisted data')
