@@ -2,7 +2,7 @@
 import abc
 import logging
 
-from typing import Dict, Optional, Set, Tuple
+from typing import Dict, List, Optional, Set, Tuple, Union
 
 from qtoggleserver.core import events as core_events
 from qtoggleserver.core import expressions as core_expressions
@@ -18,7 +18,7 @@ logger = logging.getLogger(__package__)
 class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
     logger = logger
 
-    def __init__(self, filter: dict = None) -> None:
+    def __init__(self, *, filter: dict = None, name: Optional[str] = None) -> None:
         self._filter: dict = filter or {}
         self._filter_prepared: bool = False
 
@@ -28,22 +28,24 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
         self._filter_device_attr_transitions: Dict[str, Tuple[Attribute, Attribute]] = {}
         self._filter_device_attr_names: Set[str] = set()
 
-        self._filter_port_value: Optional[NullablePortValue] = None
+        self._filter_port_value: Optional[Union[List[NullablePortValue]], core_expressions.Expression] = None
         self._filter_port_value_transition: Optional[Tuple[NullablePortValue, NullablePortValue]] = None
-        self._filter_port_attrs: Attributes = {}
+        self._filter_port_attrs: Union[Attributes, Dict[str, List[Attribute]]] = {}
         self._filter_port_attr_transitions: Dict[str, Tuple[Attribute, Attribute]] = {}
         self._filter_port_attr_names: Set[str] = set()
 
-        self._filter_slave_attrs: Attributes = {}
+        self._filter_slave_attrs: Union[Attributes, Dict[str, List[Attribute]]] = {}
         self._filter_slave_attr_transitions: Dict[str, Tuple[Attribute, Attribute]] = {}
         self._filter_slave_attr_names: Set[str] = set()
 
         # Maintain an internal "last" state for all objects, so we can detect changes in attributes and values
-        self._device_attrs: Attributes = {}
+        self._device_attrs: Union[Attributes, Dict[str, List[Attribute]]] = {}
         self._port_values: Dict[str, NullablePortValue] = {}
         self._port_attrs: Dict[str, Attributes] = {}
         self._slave_attrs: Dict[str, Attributes] = {}
 
+        super().__init__(name=name)
+
     def _prepare_filter(self) -> None:
         event_types = self._filter.get('type')
         if event_types:
@@ -55,15 +57,18 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
         if port_value is not None:
             if isinstance(port_value, str):  # An expression
                 try:
-                    self.logger.debug('using value expression "%s"', port_value)
+                    self.debug('using value expression "%s"', port_value)
                     self._filter_port_value = core_expressions.parse(self_port_id=None, sexpression=port_value)
 
-                except core_expressions.ExpressionError as e:
-                    self.logger.error('failed to parse port expression "%s": %s', port_value, e)
+                except core_expressions.ExpressionParseError as e:
+                    self.error('failed to parse port expression "%s": %s', port_value, e)
 
                     raise
 
             else:
+                if not isinstance(port_value, list):
+                    port_value = (port_value,)
+
                 self._filter_port_value = port_value
 
         self._filter_port_value_transition = self._filter.get('port_value_transition')
@@ -108,6 +113,7 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
         self._filter_slave_attr_names.update(self._filter_slave_attr_transitions.keys())
 
         self._filter_prepared = True
+        self.debug('filter prepared')
 
     @staticmethod
     def _make_changed_added_removed(old_attrs: Attributes, new_attrs: Attributes) -> Tuple[
@@ -266,9 +272,6 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
                 if new_value != self._filter_port_value.eval():
                     return False
 
-            elif new_value != self._filter_port_value:  # A single value
-                return False
-
         return True
 
     async def accepts_port(
@@ -354,16 +357,15 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
         )
 
         if not accepted:
-            self.logger.debug('skipping event %s', event)
             return
 
-        self.logger.debug('handling event %s', event)
+        self.debug('handling event %s', event)
 
         try:
             await self.on_event(event)
 
         except Exception as e:
-            self.logger.error('failed to handle event %s: %s', event, e, exc_info=True)
+            self.error('failed to handle event %s: %s', event, e, exc_info=True)
 
         try:
             if isinstance(event, core_events.ValueChange):
@@ -390,6 +392,9 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
             elif isinstance(event, core_events.DeviceUpdate):
                 await self.on_device_update(event, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs)
 
+            elif isinstance(event, core_events.FullUpdate):
+                await self.on_full_update(event)
+
             elif isinstance(event, slaves_events.SlaveDeviceUpdate):
                 await self.on_slave_device_update(
                     event,
@@ -408,7 +413,7 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
                 await self.on_slave_device_remove(event, event.get_slave(), new_attrs)
 
         except Exception as e:
-            self.logger.error('failed to handle event %s: %s', event, e, exc_info=True)
+            self.error('failed to handle event %s: %s', event, e, exc_info=True)
 
     async def on_event(self, event: core_events.Event) -> None:
         pass
@@ -455,6 +460,9 @@ class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
 
         pass
 
+    async def on_full_update(self, event: core_events.Event) -> None:
+        pass
+
     async def on_slave_device_update(
         self,
         event: core_events.Event,
