@@ -16,6 +16,7 @@ from tornado.httpclient import AsyncHTTPClient, HTTPRequest
 from qtoggleserver import persist
 from qtoggleserver.core import api as core_api
 from qtoggleserver.core import events as core_events
+from qtoggleserver.core import main as core_main
 from qtoggleserver.core import ports as core_ports
 from qtoggleserver.core import responses as core_responses
 from qtoggleserver.conf import settings
@@ -32,10 +33,13 @@ from . import exceptions
 from .ports import SlavePort
 
 
-_INVALID_EXPRESSION_RE = re.compile(r'^invalid field: ((device_)*expression)$')
+_MAX_PARALLEL_API_CALLS = 2
+_MAX_QUEUED_API_CALLS = 256
+_INVALID_EXPRESSION_FIELD_RE = re.compile(r'^((device_)*expression)$')
 _FWUPDATE_POLL_INTERVAL = 30
 _FWUPDATE_POLL_TIMEOUT = 300
 _NO_EVENT_DEVICE_ATTRS = ['uptime', 'date']
+_DEFAULT_POLL_INTERVAL = 10
 
 
 _slaves_by_name: Dict[str, Slave] = {}
@@ -52,8 +56,8 @@ class Slave(logging_utils.LoggableMixin):
         host: str,
         port: int,
         path: str,
-        poll_interval: Optional[int] = None,
-        listen_enabled: bool = True,
+        poll_interval: int = 0,
+        listen_enabled: bool = False,
         admin_password: Optional[str] = None,
         admin_password_hash: Optional[str] = None,
         last_sync: int = -1,
@@ -77,7 +81,7 @@ class Slave(logging_utils.LoggableMixin):
         self._host: str = host
         self._port: int = port
         self._path: str = path
-        self._poll_interval: Optional[int] = poll_interval
+        self._poll_interval: int = poll_interval
         self._listen_enabled: bool = listen_enabled
 
         if admin_password is not None:
@@ -101,13 +105,16 @@ class Slave(logging_utils.LoggableMixin):
         # Indicates whether all data required for this slave is present locally
         self._ready: bool = False
 
+        # API call throttling
+        self._parallel_api_caller = asyncio_utils.ParallelCaller(_MAX_PARALLEL_API_CALLS, _MAX_QUEUED_API_CALLS)
+
         # Indicates the listening session id, or None if no listen client is active
         self._listen_session_id: Optional[str] = None
         self._listen_task: Optional[asyncio.Task] = None
 
         # Tells the polling mechanism status
         self._poll_started: bool = False
-        self._pool_task: Optional[asyncio.Task] = None
+        self._poll_task: Optional[asyncio.Task] = None
 
         # Attributes cache
         self._cached_attrs: Attributes = attrs or {}
@@ -187,7 +194,7 @@ class Slave(logging_utils.LoggableMixin):
                         await self.disable()
 
                         # We have to trigger an update event here, to inform consumers about disabling
-                        self.trigger_update()
+                        await self.trigger_update()
 
                     # Check for duplicate name
                     if name in _slaves_by_name or core_device_attrs.name == name:
@@ -282,9 +289,12 @@ class Slave(logging_utils.LoggableMixin):
         elif self._listen_enabled:
             self._start_listening()
 
-        # We need to load ports asynchronously, because we expect callers of enable() to generate slave-device-update
-        # events, and we want any port-related events generated by _load_ports() to come after
-        if self._name:
+        # For permanently offline devices, we can't do an initial port discovery. We will need to load ports from
+        # persisted data.
+        #
+        # We can't await for the _load_ports() result, because we expect callers of enable() to generate
+        # slave-device-update events, and we want any port-related events generated by _load_ports() to come after.
+        if self.is_permanently_offline() and self._name:
             asyncio.create_task(self._load_ports())
 
     async def disable(self) -> None:
@@ -425,7 +435,7 @@ class Slave(logging_utils.LoggableMixin):
         # Stop polling
         if self._poll_started:
             self._stop_polling()
-            await self._pool_task
+            await self._poll_task
             self.debug('polling mechanism stopped')
 
         # Stop fwupdate pool loop
@@ -433,12 +443,15 @@ class Slave(logging_utils.LoggableMixin):
             self._fwupdate_poll_task.cancel()
             await self._fwupdate_poll_task
 
+        # Stop parallel API caller
+        self._parallel_api_caller.stop()
+
     async def _load_ports(self) -> None:
         self.debug('loading persisted ports')
         port_data_list = persist.query(SlavePort.PERSIST_COLLECTION, fields=['id'])
         my_port_ids = [
             d['id'][len(self._name) + 1:] for d in port_data_list
-            if d['id'].startswith(self._name + '.')
+            if d['id'].startswith(f'{self._name}.')
         ]
 
         for _id in my_port_ids:
@@ -455,7 +468,7 @@ class Slave(logging_utils.LoggableMixin):
         port_data_list = persist.query(SlavePort.PERSIST_COLLECTION, fields=['id'])
         my_port_data_list = [
             d for d in port_data_list
-            if d['id'].startswith(self._name + '.')
+            if d['id'].startswith(f'{self._name}.')
         ]
 
         # Remove old records
@@ -477,19 +490,25 @@ class Slave(logging_utils.LoggableMixin):
         for port in self._get_local_ports():
             await port.remove()
 
+        # Also remove persisted port data belonging to this device for ports that are no longer present on this slave
+        port_data_list = persist.query(SlavePort.PERSIST_COLLECTION, fields=['id'])
+        for d in port_data_list:
+            if d['id'].startswith(f'{self._name}.'):
+                persist.remove(SlavePort.PERSIST_COLLECTION, {'id': d['id']})
+
         self.debug('removing device')
         persist.remove('slaves', filt={'id': self._name})
 
-        self.trigger_remove()
+        await self.trigger_remove()
 
-    def trigger_add(self) -> None:
-        core_events.handle_event(events.SlaveDeviceAdd(self))
+    async def trigger_add(self) -> None:
+        await core_events.handle_event(events.SlaveDeviceAdd(self))
 
-    def trigger_remove(self) -> None:
-        core_events.handle_event(events.SlaveDeviceRemove(self))
+    async def trigger_remove(self) -> None:
+        await core_events.handle_event(events.SlaveDeviceRemove(self))
 
-    def trigger_update(self) -> None:
-        core_events.handle_event(events.SlaveDeviceUpdate(self))
+    async def trigger_update(self) -> None:
+        await core_events.handle_event(events.SlaveDeviceUpdate(self))
 
     async def api_call(
         self,
@@ -500,13 +519,24 @@ class Slave(logging_utils.LoggableMixin):
         retry_counter: Optional[int] = 0
     ) -> Any:
 
+        return await self._parallel_api_caller.call(self._api_call, method, path, body, timeout, retry_counter)
+
+    async def _api_call(
+        self,
+        method: str,
+        path: str,
+        body: Any = None,
+        timeout: int = None,
+        retry_counter: Optional[int] = 0
+    ) -> Any:
+
         if method == 'GET':
             body = None
 
         url = self.get_url(path)
         body_str = json_utils.dumps(body) if body is not None else None
 
-        # A new API call cancels any pending retry
+        # Used to signal a new API call, which should prevent any pending retry
         ref = self._last_api_call_ref = {}
 
         http_client = AsyncHTTPClient()
@@ -530,7 +560,7 @@ class Slave(logging_utils.LoggableMixin):
             request_timeout=timeout
         )
 
-        self.debug('calling api function %s %s', method, path)
+        self.debug('calling API function %s %s', method, path)
 
         try:
             response = await http_client.fetch(request, raise_error=False)
@@ -547,8 +577,8 @@ class Slave(logging_utils.LoggableMixin):
 
             msg = f'api call {method} {path} on {self} failed: {e} (body={body_str or ""})'
 
-            if (retry_counter is not None and retry_counter < settings.slaves.retry_count and
-                self._enabled and ref is not self._last_api_call_ref):
+            if ((retry_counter is not None) and (retry_counter < settings.slaves.retry_count) and
+                (ref is not self._last_api_call_ref) and self._enabled):
 
                 msg += f', retrying in {settings.slaves.retry_interval} seconds'
                 self.error(msg)
@@ -603,7 +633,7 @@ class Slave(logging_utils.LoggableMixin):
 
         self.debug('starting polling mechanism')
 
-        self._pool_task = asyncio.create_task(self._poll_loop())
+        self._poll_task = asyncio.create_task(self._poll_loop())
 
     def _stop_polling(self) -> None:
         if not self._poll_started:
@@ -614,6 +644,7 @@ class Slave(logging_utils.LoggableMixin):
         self.debug('stopping polling mechanism')
 
         self._poll_started = False
+        self._poll_task.cancel()
 
     def _start_fwupdate_polling(self) -> None:
         if self._fwupdate_poll_task:
@@ -715,7 +746,7 @@ class Slave(logging_utils.LoggableMixin):
     def _get_local_ports(self) -> List[SlavePort]:
         return [
             port for port in core_ports.all_ports()
-            if port.get_id().startswith(self._name + '.') and isinstance(port, SlavePort)
+            if port.get_id().startswith(f'{self._name}.') and isinstance(port, SlavePort)
         ]
 
     async def _listen_loop(self) -> None:
@@ -753,7 +784,7 @@ class Slave(logging_utils.LoggableMixin):
                     request_timeout=settings.slaves.timeout + settings.slaves.keepalive
                 )
 
-                self.debug('calling api function GET /listen')
+                self.debug('calling API function GET /listen')
 
                 try:
                     response = await http_client.fetch(request, raise_error=False)
@@ -1083,7 +1114,7 @@ class Slave(logging_utils.LoggableMixin):
 
         self.debug('handling event of type %s', event['type'])
         try:
-            await method(**event['params'])
+            await method(**event.get('params', {}))
 
         except exceptions.DeviceRenamed:
             # Treat DeviceRenamed as an expected exception, do not log anything but forward it
@@ -1110,15 +1141,12 @@ class Slave(logging_utils.LoggableMixin):
             json_utils.dumps(value)
         )
 
-        # Trigger a master value-change if the returned value has not changed from the one we locally have (this happens
-        # when slave indirectly rejects pushed value, by triggering itself a value-change with the old value)
-        if port.get_cached_value() == value:
-            port.trigger_value_change()
-
         port.set_cached_value(value)
         port.update_last_sync()
         await port.save()
 
+        await core_main.update()
+
     async def _handle_port_update(self, **attrs: Attribute) -> None:
         local_id = f'{self._name}.{attrs.get("id")}'
         port = core_ports.get(local_id)
@@ -1148,12 +1176,13 @@ class Slave(logging_utils.LoggableMixin):
                 await port.handle_attr_change(name, value)
 
         port.update_cached_attrs(attrs)
+        await port.update_enabled()
 
         if 'value' in attrs:  # Value has also been updated
             port.update_last_sync()
 
         await port.save()
-        port.trigger_update()
+        await port.trigger_update()
 
     async def _handle_port_add(self, **attrs: Attribute) -> None:
         local_id = f'{self._name}.{attrs.get("id")}'
@@ -1181,18 +1210,24 @@ class Slave(logging_utils.LoggableMixin):
                 attrs.pop(name)
 
         await self.update_cached_attrs(attrs)
-        self.trigger_update()
+        await self.trigger_update()
+        self.save()
+
+    async def _handle_full_update(self) -> None:
+        await self.fetch_and_update_device()
+        await self.fetch_and_update_ports()
+        await self.trigger_update()
         self.save()
 
     async def _handle_offline(self) -> None:
         self.debug('device is offline')
 
-        self.trigger_update()
+        await self.trigger_update()
 
         # Trigger a port-update so that online attribute is pushed to consumers
         for port in self._get_local_ports():
             if port.is_enabled():
-                port.trigger_update()
+                await port.trigger_update()
 
     async def _handle_online(self) -> None:
         self.debug('device is online')
@@ -1216,12 +1251,12 @@ class Slave(logging_utils.LoggableMixin):
 
                 return
 
-            self.trigger_update()
+        await self.trigger_update()
 
         # Trigger a port-update so that online attribute is pushed to consumers
         for port in self._get_local_ports():
             if port.is_enabled():
-                port.trigger_update()
+                await port.trigger_update()
 
         if not self._ready:
             self.debug('device is ready')
@@ -1448,7 +1483,7 @@ class Slave(logging_utils.LoggableMixin):
                     self._cached_attrs[name] = value
 
                 # Inform clients about the provisioning field change
-                self.trigger_update()
+                await self.trigger_update()
 
                 self.save()
 
@@ -1484,7 +1519,7 @@ class Slave(logging_utils.LoggableMixin):
         if path == '/device':
             if method == 'PATCH':
                 # Intercept this API call to detect admin password changes
-                new_admin_password = request_body and request_body.get('admin_password') or None
+                new_admin_password = (request_body or {}).get('admin_password')
                 if new_admin_password is not None:
                     self.debug('updating admin password')
                     self.set_admin_password(new_admin_password)
@@ -1511,7 +1546,7 @@ class Slave(logging_utils.LoggableMixin):
 
                 self.debug('firmware update process active')
                 await self.disable()
-                self.trigger_update()
+                await self.trigger_update()
                 self._start_fwupdate_polling()
 
             elif method == 'GET' and self._fwupdate_poll_task:
@@ -1526,15 +1561,19 @@ class Slave(logging_utils.LoggableMixin):
 
                 self.debug('device has been reset to factory defaults')
                 await self.disable()
-                self.trigger_update()
+                await self.trigger_update()
 
     def intercept_error(self, error: Exception) -> Exception:
         if isinstance(error, core_responses.HTTPError):
             # Slave expression attribute is known as "device_expression" on Master; we must adapt the corresponding
-            # error here
-            m = _INVALID_EXPRESSION_RE.match(error.msg)
-            if m:
-                return core_responses.HTTPError(error.code, 'invalid field: device_' + m.group(1))
+            # error here by prepending a(nother) "device_"
+            if error.code == 'invalid-field':
+                field = error.params.get('field', '')
+                m = _INVALID_EXPRESSION_FIELD_RE.match(field)
+                if m:
+                    params = dict(error.params)
+                    params['field'] = 'device_' + m.group(1)
+                    return core_responses.HTTPError(error.status, error.code, **params)
 
         return error
 
@@ -1548,26 +1587,53 @@ async def add(
     host: str,
     port: int,
     path: str,
-    poll_interval: Optional[int],
-    listen_enabled: bool,
+    poll_interval: int = 0,
+    listen_enabled: Optional[bool] = None,
     admin_password: Optional[str] = None,
     admin_password_hash: Optional[str] = None
 ) -> Slave:
 
-    slave = Slave(None, scheme, host, port, path, poll_interval, listen_enabled, admin_password, admin_password_hash)
+    slave = Slave(
+        name=None,
+        scheme=scheme,
+        host=host,
+        port=port,
+        path=path,
+        poll_interval=0,  # Will be set later
+        listen_enabled=False,  # Will be set later
+        admin_password=admin_password,
+        admin_password_hash=admin_password_hash
+    )
 
     slave.debug('starting add procedure')
 
     await slave.fetch_and_update_device()
     name = slave.get_name()
 
-    if listen_enabled and 'listen' not in slave.get_cached_attr('flags'):
+    # Check that we have required listen support
+    if (listen_enabled is True) and 'listen' not in slave.get_cached_attr('flags'):
         slave.error('no listen support')
         raise exceptions.NoListenSupport(name)
 
+    # If no listen and no polling specified, attempt to automatically detect and enable supported method
+    if (listen_enabled is None) and (poll_interval == 0):
+        if 'listen' in slave.get_cached_attr('flags'):
+            slave.debug('listen support detected, auto-enabling listening')
+            listen_enabled = True
+
+        else:
+            slave.debug('listen support not detected, auto-enabling polling')
+            poll_interval = _DEFAULT_POLL_INTERVAL
+
+    if listen_enabled:
+        slave.enable_listen()
+
+    elif poll_interval:
+        slave.set_poll_interval(poll_interval)
+
     await slave.enable()
+    await slave.trigger_add()
     slave.save()
-    slave.trigger_add()
 
     if not listen_enabled and not poll_interval:
         # Device is permanently offline, but we must know its ports; this would otherwise be called by
@@ -1609,7 +1675,12 @@ async def load() -> None:
     _load_time = time.time()
 
     for entry in persist.query('slaves'):
-        entry['name'] = entry.pop('id')
+        try:
+            entry['name'] = entry.pop('id')
+
+        except KeyError:
+            logger.error('skipping entry with missing "id" key in persisted data')
+            continue
 
         try:
             slave = Slave(**entry)
@@ -1627,9 +1698,10 @@ async def load() -> None:
         else:
             logger.debug('loaded %s (disabled)', slave)
 
-        slave.trigger_add()
+        await slave.trigger_add()
 
 
 async def cleanup() -> None:
-    for slave in _slaves_by_name.values():
-        await slave.cleanup()
+    tasks = [slave.cleanup() for slave in _slaves_by_name.values()]
+    if tasks:
+        await asyncio.wait(tasks)
